/*
  jn516x_prog.h - Flash Programmer Library for NXP JN516x Series

  Copyright (c) 2022-2024 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include "ftd2.h"
#include "jn516x.h"

#define JN516X_MAX_DATA_SIZE 128
#define OTA_BOOT_FIRMWARE "-OTA-BOOT.bin"

class JN516x_PROG : public JN516x, public FTD2
{
  public:

    typedef enum
    {
      /* boot loader */
      STATUS_OK               = 0,
      STATUS_NOT_SUPPORTED    = 0xFF,
      STATUS_WRITE_FAIL       = 0xFE,
      STATUS_INVALID_RESPONSE = 0xFD,
      STATUS_CRC_ERROR        = 0xFC,
      STATUS_ASSERT_FAIL      = 0xFB,
      STATUS_USER_INTERRUPT   = 0xFA,
      STATUS_READ_FAIL        = 0xF9,
      STATUS_TST_ERROR        = 0xF8,
      STATUS_AUTH_ERROR       = 0xF7,
      STATUS_NO_RESPONSE      = 0xF6,
      /* orignal */
      STATUS_FILE_WRITE_FAIL  = 0xE3,
      STATUS_FILE_READ_FAIL   = 0xE2,
      STATUS_FILE_NOT_FOUND   = 0xE1,
      STATUS_INVALID_VERSION  = 0xE0,
    } STATUS;

    typedef struct __attribute__((packed))
    {
      uint8_t  magicNumber[12];
      uint8_t  configuration;
      uint8_t  status;
      uint16_t applicationID;
    } tsBIR;

    typedef struct __attribute__((packed))
    {
      tsBIR    sBIR;           // 16-byte Boot Image Record
      uint8_t  encryption[14]; // Encryption Initialisation Vector (ignored if unencrypted)
      uint16_t configuration;  // 16-bit Software Configuration Options
      uint32_t imageSize;      // 32-bit Length of Binary Image in bytes
      uint32_t dataStart;      // 32-bit .data section Flash start address
      uint16_t dataAddress;    // 16-bit .data section load address in RAM (word aligned)
      uint16_t dataLength;     // 16-bit .data section length in 32-bit words
      uint16_t bssStart;       // 16 bit .bss section start address in RAM (word aligned)
      uint16_t bssLength;      // 16-bit .bss section length in 32-bit words
      uint32_t warmStart;      // 32-bit wake-up entry point (word aligned) – warm start
      uint32_t coldStart;      // 32-bit reset entry point (word aligned) – cold start
    } tsFlashHeader;

    typedef struct __attribute__((packed))
    {
      uint32_t version;        // version
      tsBIR    sBIR;           // 16-byte Boot Image Record
      uint8_t  encryption[14]; // Encryption Initialisation Vector (ignored if unencrypted)
      uint16_t configuration;  // 16-bit Software Configuration Options
      uint32_t imageSize;      // 32-bit Length of Binary Image in bytes
      uint32_t dataStart;      // 32-bit .data section Flash start address
      uint16_t dataAddress;    // 16-bit .data section load address in RAM (word aligned)
      uint16_t dataLength;     // 16-bit .data section length in 32-bit words
      uint16_t bssStart;       // 16 bit .bss section start address in RAM (word aligned)
      uint16_t bssLength;      // 16-bit .bss section length in 32-bit words
      uint32_t warmStart;      // 32-bit wake-up entry point (word aligned) – warm start
      uint32_t coldStart;      // 32-bit reset entry point (word aligned) – cold start
    } tsFlashHeader_JN5179;

    JN516x_PROG(void)
    : _debug(false)
    {
    }

    virtual ~JN516x_PROG(void)
    {
    }

    static const char *errstr(STATUS status)
    {
      switch (status)
      {
        case STATUS_OK: return "Success";
        /* boot loader */
        case STATUS_NOT_SUPPORTED:    return "not supported";
        case STATUS_WRITE_FAIL:       return "write failed";
        case STATUS_INVALID_RESPONSE: return "invalid responce";
        case STATUS_CRC_ERROR:        return "CRC error";
        case STATUS_ASSERT_FAIL:      return "assert failed";
        case STATUS_USER_INTERRUPT:   return "user interrupt";
        case STATUS_READ_FAIL:        return "read failed";
        case STATUS_TST_ERROR:        return "test error";
        case STATUS_AUTH_ERROR:       return "auth error";
        case STATUS_NO_RESPONSE:      return "no responce";
        /* orignal */
        case STATUS_FILE_WRITE_FAIL:  return "file write failed";
        case STATUS_FILE_READ_FAIL:   return "file read failed";
        case STATUS_FILE_NOT_FOUND:   return "file not found";
        case STATUS_INVALID_VERSION:  return "invalid version";
      }
      return "unknown error";
    }

    void setChipIDPath(const char *path)
    {
      _chipidpath = path;
    }

    VERSION chipVersion(CHIPID id)
    {
      char keyname[16], chipver[16];
      if (_chipidpath.length())
      {
        snprintf(keyname, sizeof(keyname), "%08X", id);
        if (GetPrivateProfileString("chipid", keyname, "", chipver, sizeof(chipver), _chipidpath.c_str()))
          return (VERSION)strtoul(chipver, nullptr, 16);
      }
      VERSION version = JN516x::chipVersion(id);
      unsigned char b;
      switch (version)
      {
        case VERSION_JN517x:
          _debug = false;
          if (readFlash(160 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = (VERSION)0;
          else if (readFlash(256 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = VERSION_JN5174;
          else if (readFlash(512 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = VERSION_JN5178;
          else
          {
            version = VERSION_JN5179;
            break;
          }
          /* restart boot mode */
          program(_speed, _timeout);
          break;
        case VERSION_JN516x:
          _debug = false;
          if (readFlash( 64 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = (VERSION)0;
          else if (readFlash(160 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = VERSION_JN5161;
          else if (readFlash(256 * 1024 - 1, &b, sizeof(b)) != STATUS_OK)
            version = VERSION_JN5164;
          else
          {
            version = VERSION_JN5168;
            break;
          }
          /* restart boot mode */
          program(_speed, _timeout);
          break;
        default: break;
      }
      if (_chipidpath.length() && version)
      {
        snprintf(chipver, sizeof(chipver), "%08X", version);
        WritePrivateProfileString("chipid", keyname, chipver, _chipidpath.c_str());
      }
      return version;
    }

    STATUS eraseDevice(void)
    {
      CHIPID chipid;
      STATUS rv = chipID(chipid);
      if (rv == STATUS_OK)
      {
        FLASHID flashid;
        if ((rv = flashID(flashid)) == STATUS_OK)
        {
          if ((rv = selectFlash(flashType(flashid))) == STATUS_OK)
          {
            if ((rv = writeSR(0)) == STATUS_OK)
              rv = eraseFlash();
          }
        }
      }
      return rv;
    }

    STATUS writeDevice(const char *filename, uint8_t u8Sector = 0)
    {
      CHIPID  chipid;
      VERSION chipver;
      STATUS rv = chipID(chipid);
      chipver = chipVersion(chipid);
      programStart(chipName(chipver), filename, true);
      if (rv == STATUS_OK)
      {
        rv = STATUS_FILE_NOT_FOUND;
        FILE *fp = fopen(filename, "rb");
        if (fp)
        {
          VERSION filever;
          if ((rv = fileVersion(filever, fp)) == STATUS_OK)
          {
            if (chipType(filever) != chipType(chipver))
              rv = STATUS_INVALID_VERSION;
            else
            {
              FLASHID flashid;
              if ((rv = flashID(flashid)) == STATUS_OK)
              {
                if ((rv = selectFlash(flashType(flashid))) == STATUS_OK)
                {
                  if ((rv = writeSR(0)) == STATUS_OK)
                  {
                    uint8_t data[JN516X_MAX_DATA_SIZE];
                    size_t fpos = (chipver == VERSION_JN5179 ? 0 : ftell(fp));
                    fseek(fp, 0, SEEK_END);
                    size_t  size  = ftell(fp) - fpos;
                    if (strstr(filename, OTA_BOOT_FIRMWARE))
                      u8Sector = chipFlashSectors(chipver) - 1;
                    size_t  offs  = FLASH_SECTOR_OFFSET(u8Sector);
                    size_t  addr  = 0;
                    size_t  prog  = 0;
                    uint8_t erase = 0xFF;
                    fseek(fp, fpos, SEEK_SET);
                    while (1)
                    {
                      ssize_t n = fread(data, 1, sizeof(data), fp);
                        if (n <= 0)
                          break;
                      u8Sector = (addr + offs) >> 15;
                      if (erase != u8Sector)
                      {
                        erase = u8Sector;
                        if ((rv = eraseFlash(u8Sector)))
                          break;
                      }
                      if ((rv = programFlash(addr + offs, data, n)) != STATUS_OK)
                        break;
                      addr += n;
                      if (prog != (addr >> 10))
                      {
                        prog = (addr >> 10);
                        programProgress(size, addr);
                      }
                    }
                    programProgress(size, addr);
                  }
                }
              }
            }
          }
          fclose(fp);
        }
      }
      programEnd(rv);
      return rv;
    }

    STATUS readDevice(const char *filename, uint8_t u8Sector = 0)
    {
      CHIPID  chipid;
      VERSION chipver;
      STATUS rv = chipID(chipid);
      chipver = chipVersion(chipid);
      programStart(chipName(chipver), filename, false);
      if (rv == STATUS_OK)
      {
        rv = STATUS_FILE_NOT_FOUND;
        FILE *fp = fopen(filename, "wb");
        if (fp)
        {
          FLASHID flashid;
          if ((rv = flashID(flashid)) == STATUS_OK)
          {
            if ((rv = selectFlash(flashType(flashid))) == STATUS_OK)
            {
              if ((rv = writeSR(0)) == STATUS_OK)
              {
                /* Search Flash Header */
                size_t  offs = FLASH_SECTOR_OFFSET(u8Sector);
                size_t  size = chipFlashSize(chipver);
                bool    boot;
                while (offs < size)
                {
                  boot = false;
                  if (chipver == VERSION_JN5179)
                  {
                    tsFlashHeader_JN5179 sFlashHeader;
                    if ((rv = readFlash(offs, (uint8_t *)&sFlashHeader, sizeof(sFlashHeader))) != STATUS_OK)
                      break;
                    boot |= memcmp(sFlashHeader.sBIR.magicNumber, JN516x::MAGICNUMBER, sizeof(JN516x::MAGICNUMBER)) == 0;
                    boot |= memcmp(sFlashHeader.sBIR.magicNumber, JN516x::MAGICNUMBER_JN5179, sizeof(JN516x::MAGICNUMBER)) == 0;
                    boot &= sFlashHeader.sBIR.status == 0x01;
                    boot &= sFlashHeader.sBIR.applicationID == 0x0000;
                    if (boot)
                    {
                      size = read32((uint8_t *)&sFlashHeader.imageSize);
                      break;
                    }
                  }
                  else
                  {
                    tsFlashHeader sFlashHeader;
                    if ((rv = readFlash(offs, (uint8_t *)&sFlashHeader, sizeof(sFlashHeader))) != STATUS_OK)
                      break;
                    boot |= memcmp(sFlashHeader.sBIR.magicNumber, JN516x::MAGICNUMBER, sizeof(JN516x::MAGICNUMBER)) == 0;
                    boot &= sFlashHeader.sBIR.status == 0x01;
                    boot &= sFlashHeader.sBIR.applicationID == 0x0000;
                    if (boot)
                    {
                      size = read32((uint8_t *)&sFlashHeader.imageSize);
                      break;
                    }
                  }
                  offs += 0x8000U;
                }
                if (boot)
                {
                  uint8_t data[JN516X_MAX_DATA_SIZE];
                  if (chipver != VERSION_JN5179)
                  {
                    write32(data, chipver, false);
                    if (fwrite(data, 1, 4, fp) != 4)
                      rv = STATUS_FILE_WRITE_FAIL;
                  }
                  if (rv == STATUS_OK)
                  {
                    size_t addr = 0;
                    size_t prog = 0;
                    while (addr < size)
                    {
                      size_t n = __min(size - addr, sizeof(data));
                      if ((rv = readFlash(addr + offs, data, n)) != STATUS_OK)
                        break;
                      if (fwrite(data, 1, n, fp) != n)
                      {
                        rv = STATUS_FILE_WRITE_FAIL;
                        break; // write error
                      }
                      addr += n;
                      if (prog != (addr >> 10))
                      {
                        prog = (addr >> 10);
                        programProgress(size, addr);
                      }
                    }
                    programProgress(size, addr);
                  }
                }
              }
            }
          }
          fclose(fp);
        }
      }
      programEnd(rv);
      return rv;
    }

    STATUS program(SPEED speed = SPEED_1M, uint32_t timeout = 1000)
    {
      // setup uart
      init(SPEED_38400, MODE_8N1, FLOW_RTS_CTS, timeout);
      // setup cbus
      digitalWrite(PIN_PROG , HIGH);
      digitalWrite(PIN_RESET, HIGH);
      pinMode(PIN_PROG , GPIO_OUTPUT);
      pinMode(PIN_RESET, GPIO_OUTPUT);
      // start program mode
      digitalWrite(PIN_PROG, LOW);
      restart(true);
      Sleep::sleep(300);
      digitalWrite(PIN_PROG, HIGH);
      // change baudrate
      STATUS status = changeBaudrate(speed);
      if (status == STATUS_OK)
      {
        _debug   = true;
        _speed   = speed;
        _timeout = timeout;
      }
      return status;
    }

    STATUS restart(bool hardware = true)
    {
      if (hardware)
      {
        digitalWrite(PIN_RESET, LOW);
        digitalWrite(PIN_RESET, HIGH);
        return STATUS_OK;
      }
      return protocol(MSGTYPE_SET_RESET_REQUEST);
    }

    STATUS changeBaudrate(SPEED speed)
    {
      uint8_t arg;
      switch (speed)
      {
        case SPEED_1M    : arg =  1; break;
        case SPEED_500K  : arg =  2; break;
        case SPEED_115200: arg =  9; break;
        case SPEED_38400 : arg = 26; break;
        default: return STATUS_NOT_SUPPORTED;
      }
      STATUS rv = protocol(MSGTYPE_CHANGE_BAUDRATE_REQUEST, &arg, sizeof(arg));
      if (rv == STATUS_OK)
        baudrate(speed);
      return rv;
    }

    STATUS eraseFlash(void)
    {
      return protocol(MSGTYPE_FLASH_ERASE_REQUEST);
    }

    STATUS eraseFlash(uint8_t sector)
    {
      return protocol(MSGTYPE_SECTOR_ERASE_REQUEST, &sector, sizeof(sector));
    }

    STATUS programFlash(uint32_t addr, uint8_t *buf, uint16_t len)
    {
      uint8_t data[JN516X_MAX_DATA_SIZE + 4];
      if (len > JN516X_MAX_DATA_SIZE)
        len = JN516X_MAX_DATA_SIZE;
      write32(data + 0, addr);
      memcpy(data + 4, buf, len);
      return protocol(MSGTYPE_FLASH_PROGRAM_REQUEST, data, len + 4);
    }

    STATUS readFlash(uint32_t addr, uint8_t *buf, uint16_t len)
    {
      if (len > JN516X_MAX_DATA_SIZE)
        len = JN516X_MAX_DATA_SIZE;
      uint8_t data[6];
      write32(data + 0, addr);
      write16(data + 4, len);
      STATUS rv = protocol(MSGTYPE_FLASH_READ_REQUEST, data, sizeof(data));
      if (rv == STATUS_OK)
        memcpy(buf, _packet + 3, len);
      return rv;
    }

    STATUS selectFlash(FLASH type, uint32_t jump = 0)
    {
      uint8_t data[5];
      data[0] = (uint8_t)type;
      write32(data + 1, jump);
      return protocol(MSGTYPE_SELECT_FLASH_TYPE_REQUEST, data, sizeof(data));
    }

    STATUS writeSR(uint8_t data)
    {
      return protocol(MSGTYPE_WRITE_SR_REQUEST, &data, sizeof(data));
    }

    STATUS writeRAM(uint32_t addr, uint8_t *buf, uint16_t len)
    {
      uint8_t data[JN516X_MAX_DATA_SIZE + 4];
      if (len > JN516X_MAX_DATA_SIZE)
        len = JN516X_MAX_DATA_SIZE;
      write32(data + 0, addr);
      memcpy(data + 4, buf, len);
      return protocol(MSGTYPE_RAM_WRITE_REQUEST, data, len + 4);
    }

    STATUS readRAM(uint32_t addr, uint8_t *buf, uint16_t len)
    {
      if (len > JN516X_MAX_DATA_SIZE)
        len = JN516X_MAX_DATA_SIZE;
      uint8_t data[6];
      write32(data + 0, addr);
      write16(data + 4, len);
      STATUS rv = protocol(MSGTYPE_RAM_READ_REQUEST, data, sizeof(data));
      if (rv == STATUS_OK)
        memcpy(buf, _packet + 3, len);
      return rv;
    }

    STATUS run(uint32_t jump)
    {
      uint8_t data[4];
      write32(data, jump);
      return protocol(MSGTYPE_RUN_REQUEST, data, sizeof(data));
    }

    STATUS chipID(CHIPID& id)
    {
      STATUS rv = protocol(MSGTYPE_GET_CHIP_ID_REQUEST);
      id = (CHIPID)(rv == STATUS_OK ? read32(_packet + 3) : 0);
      return rv;
    }

    STATUS flashID(FLASHID& id)
    {
      STATUS rv = protocol(MSGTYPE_READ_FLASH_ID_REQUEST);
      id = (FLASHID)(rv == STATUS_OK ? read16(_packet + 3) : 0);
      return rv;
    }

    FLASH flashType(FLASHID id)
    {
      switch (id)
      {
        case FLASHID_ST_M25P05_A:    return FLASH_ST_M25P05_A;
        case FLASHID_ST_M25P10_A:    return FLASH_ST_M25P10_A;
        case FLASHID_ST_M25P20_A:    return FLASH_ST_M25P20_A;
        case FLASHID_ST_M25P40:      return FLASH_ST_M25P40;
        case FLASHID_SST_25VF010A:   return FLASH_SST_25VF010A;
        case FLASHID_ATMEL_25F512:   return FLASH_ATMEL_25F512;
        case FLASHID_INTERNAL_FLASH: return FLASH_INTERNAL_FLASH;
        default: break;
      }
      return (FLASH)0;
    }

    STATUS bootVersion(uint32_t& version)
    {
      uint8_t buf[sizeof(version)];
      STATUS rv = readRAM(0x0004, buf, sizeof(buf));
      version = (rv == STATUS_OK ? read32(buf) : 0);
      return rv;
    }

    STATUS macAddress(uint8_t *buf, bool customer = true)
    {
      return readRAM(customer ? 0x01001570 : 0x01001580, buf, 8);
    }

  protected:

    virtual void programStart(const char *chipname, const char *filename, bool program) {}
    virtual void programProgress(size_t size, size_t processed) {}
    virtual void programEnd(STATUS status) {}

  private:

    //
    // for MONOWIRELESS TWELITE-R(2) / MONOSTICK
    //
    static const CBUS PIN_PROG  = CBUS_3;
    static const CBUS PIN_RESET = CBUS_2;

    typedef enum : uint8_t
    {
      MSGTYPE_FLASH_ERASE_REQUEST        = 0x07,
      MSGTYPE_FLASH_ERASE_RESPONSE       = 0x08,
      MSGTYPE_FLASH_PROGRAM_REQUEST      = 0x09,
      MSGTYPE_FLASH_PROGRAM_RESPONSE     = 0x0A,
      MSGTYPE_FLASH_READ_REQUEST         = 0x0B,
      MSGTYPE_FLASH_READ_RESPONSE        = 0x0C,
      MSGTYPE_SECTOR_ERASE_REQUEST       = 0x0D,
      MSGTYPE_SECTOR_ERASE_RESPONSE      = 0x0E,
      MSGTYPE_WRITE_SR_REQUEST           = 0x0F,
      MSGTYPE_WRITE_SR_RESPONSE          = 0x10,
      MSGTYPE_SET_RESET_REQUEST          = 0x14,
      MSGTYPE_SET_RESET_RESPONSE         = 0x15,
      MSGTYPE_RAM_WRITE_REQUEST          = 0x1D,
      MSGTYPE_RAM_WRITE_RESPONSE         = 0x1E,
      MSGTYPE_RAM_READ_REQUEST           = 0x1F,
      MSGTYPE_RAM_READ_RESPONSE          = 0x20,
      MSGTYPE_RUN_REQUEST                = 0x21,
      MSGTYPE_RUN_RESPONSE               = 0x22,
      MSGTYPE_READ_FLASH_ID_REQUEST      = 0x25,
      MSGTYPE_READ_FLASH_ID_RESPONSE     = 0x26,
      MSGTYPE_CHANGE_BAUDRATE_REQUEST    = 0x27,
      MSGTYPE_CHANGE_BAUDRATE_RESPONSE   = 0x28,
      MSGTYPE_SELECT_FLASH_TYPE_REQUEST  = 0x2C,
      MSGTYPE_SELECT_FLASH_TYPE_RESPONSE = 0x2D,
      MSGTYPE_GET_CHIP_ID_REQUEST        = 0x32,
      MSGTYPE_GET_CHIP_ID_RESPONSE       = 0x33,
    } MSGTYPE;

    void write32(uint8_t *buf, uint32_t data, bool littleEndian = true)
    {
      if (littleEndian)
      {
        buf[0] = (uint8_t)data;
        buf[1] = (uint8_t)(data >> 8);
        buf[2] = (uint8_t)(data >> 16);
        buf[3] = (uint8_t)(data >> 24);
      }
      else
      {
        buf[3] = (uint8_t)data;
        buf[2] = (uint8_t)(data >> 8);
        buf[1] = (uint8_t)(data >> 16);
        buf[0] = (uint8_t)(data >> 24);
      }
    }

    void write16(uint8_t *buf, uint16_t data, bool littleEndian = true)
    {
      if (littleEndian)
      {
        buf[0] = (uint8_t)data;
        buf[1] = (uint8_t)(data >> 8);
      }
      else
      {
        buf[1] = (uint8_t)data;
        buf[0] = (uint8_t)(data >> 8);
      }
    }

    uint32_t read32(const uint8_t *buf, bool littleEndian = false)
    {
      return littleEndian
        ? ((uint32_t)buf[3] << 24) |
          ((uint32_t)buf[2] << 16) |
          ((uint16_t)buf[1] <<  8) |
          buf[0]
        : ((uint32_t)buf[0] << 24) |
          ((uint32_t)buf[1] << 16) |
          ((uint16_t)buf[2] <<  8) |
          buf[3];
    }

    uint16_t read16(const uint8_t *buf, bool littleEndian = false)
    {
      return littleEndian
        ? ((uint16_t)buf[1] << 8) | buf[0]
        : ((uint16_t)buf[0] << 8) | buf[1];
    }

    STATUS fileVersion(VERSION& version, FILE *fp)
    {
      uint8_t data[4];
      fseek(fp, 0, SEEK_SET);
      if (fread(data, 1, sizeof(data), fp) != sizeof(data))
      {
        version = (VERSION)0;
        return STATUS_FILE_READ_FAIL;
      }
      version = (VERSION)read32(data);
      return STATUS_OK;
    }

    uint8_t csum(void *buf, size_t len)
    {
      uint8_t *p = (uint8_t *)buf;
      uint8_t  c = 0;
      while (len--)
        c ^= *p++;
      return c;
    }

    STATUS protocol(MSGTYPE msg, uint8_t *buf = nullptr, uint8_t len = 0)
    {
      memcpy(_packet + 2, buf, len);
      len += 2;
      _packet[0]   = len;
      _packet[1]   = msg;
      _packet[len] = csum(_packet, len);
      len += 1;
      if (write(_packet, len) != len)
        return STATUS_WRITE_FAIL;
      for (len = _packet[0] = 0; len <= _packet[0]; )
      {
        uint8_t n = read(_packet + len, _packet[0] - len + 1);
        if (n == 0)
        {
          if (_debug)
            printf("Error: Message Type [%02X]\n", msg);
          return STATUS_NO_RESPONSE;
        }
        len += n;
      }
      return csum(_packet, len) ? STATUS_CRC_ERROR : (STATUS)_packet[2];
    }

    bool        _debug;
    SPEED       _speed;
    uint32_t    _timeout;
    std::string _chipidpath;
    uint8_t     _packet[256];
};
