/*
 main.cpp - Flash Programmer Utility for NXP JN516x Series

 Copyright (c) 2022-2024 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <limits.h>
#include "jn516x_prog.h"

class Device : public JN516x_PROG
{
  public:

    Device(void) {}
    virtual ~Device(void) {}

  protected:

    void programStart(const char *chipname, const char *filename, bool program) override
    {
      if (program)
        printf("Write Device [%s] <- \"%s\"\n", chipname, filename);
      else
        printf("Read Device [%s] -> \"%s\"\n", chipname, filename);
    }

    void programProgress(size_t size, size_t processed) override
    {
      printf("Progress: %6llu/%6llu\r", processed, size);
    }

    void programEnd(STATUS status) override
    {
      printf("\n");
      if (status != STATUS_OK)
        printf("Error: [0x%02X] ", status);
      printf("%s\n", errstr(status));
    }
};

static Device device;

void printMAC(const char *message, bool customer)
{
  uint8_t buf[8];
  device.macAddress(buf, customer);
  printf("%s", message);
  for (int i = 0; i < 8; ++i)
  {
    if (i)
      printf(":");
    printf("%02X", buf[i]);
  }
  printf("\n");
}

void chipInfo(void)
{
  Device::CHIPID  chipid;
  Device::VERSION chipver;
  Device::FLASHID flashid;
  uint32_t bootver;
  device.chipID(chipid);
  chipver = device.chipVersion(chipid);
  device.flashID(flashid);
  device.bootVersion(bootver);
  printf("BootLoader Version: 0x%08X\n", bootver);
  printf("Flash ID: 0x%04X, Type: %u\n", flashid, device.flashType(flashid));
  printf("Chip  ID: 0x%08X, Name: %s", chipid, device.chipName(chipver));
  printf(", Flash: %u Kb, Ram: %u Kb, Type: %u\n", device.chipFlashSize(chipver) / 1024, device.chipRamSize(chipver) / 1024, device.chipType(chipver));
  printMAC("Default  MAC Address: ", false);
  printMAC("Customer MAC Address: ", true );
}

void portScan(void)
{
  int count = device.getDeviceCount();
  printf("Device Count: %d\n", count);
  for (int i = 0; i < count; ++i)
  {
    char buf[128];
    printf("Device-%d: ", i);
    if (device.begin(i))
    {
      printf("Port: Com%u", device.getComPortNumber());
      device.end();
      device.getDeviceSerial(i, buf);
      printf(", Serial: %s", buf);
      device.getDeviceDescription(i, buf);
      printf(", Description: %s", buf);
    }
    else
      printf("-- Open failure --");
    printf("\n");
  }
}

void usage(bool help)
{
  printf("Flash Programmer Utility for NXP JN516X, Version 1.1\n");
  printf("Copyright(c) 2022-2024 Sasapea's Lab. All right reserved.\n");
  printf("\n");
  if (help)
  {
    printf("Usage: JN516X-PROG [options]\n");
    printf("\n");
    printf("  options:\n");
    printf("    -s          ... FTD2XX Port Scan\n");
    printf("    -p portname ... FTD2XX Serial Number for Open\n");
    printf("    -b          ... baudrate (38400,115200,500K,*1M)\n");
    printf("    -d          ... device information\n");
    printf("    -e          ... erase device\n");
    printf("    -w filename ... write device (.bin)\n");
    printf("    -r filename ... read  device (.bin)\n");
    printf("    -f sector   ... start sector (*0 - n)\n");
    printf("    -h          ... this help\n");
    printf("\n");
  }
}

int main(int argc, char **argv)
{
  bool help  = false;
  bool scan  = false;
  bool info  = false;
  bool erase = false;
  const char *portname = "";
  const char *firmware = "";
  const char *savefile = "";
  Device::SPEED speed  = Device::SPEED_1M;
  uint8_t sector = 0;

  for (int i = 1; i < argc; )
  {
    char *p = argv[i++];
    if (*p == '-')
    {
      switch (*++p)
      {
        case 'h':
          help = true;
          break;
        case 's':
          scan = true;
          break;
        case 'p':
          if (i < argc)
            portname = argv[i++];
          break;
        case 'w':
          if (i < argc)
            firmware = argv[i++];
          break;
        case 'r':
          if (i < argc)
            savefile = argv[i++];
          break;
        case 'b':
          if (i < argc)
          {
            switch (strtoul(argv[i++], &p, 10))
            {
              case 38400:
                speed = Device::SPEED_38400;
                break;
              case 115200:
                speed = Device::SPEED_115200;
                break;
              case 500:
                if (toupper(*p) != 'K')
                  break;
                /* Fall Through */
              case 500000:
                speed = Device::SPEED_500K;
                break;
              case 1:
                if (toupper(*p) != 'M')
                  break;
                /* Fall Through */
              case 1000000:
                speed = Device::SPEED_1M;
                break;
            }
          }
          break;
        case 'f':
          if (i < argc)
            sector = atoi(argv[i++]);
          break;
        case 'd':
          info = true;
          break;
        case 'e':
          erase = true;
          break;
        default:
          help = true;
          break;
      }
    }
    else
      help = true;
  }

  if (help)
    usage(true);
  else if (device.status() != FT_OK)
  {
    usage(false);
    printf("FTD2XX Library Not Found.");
  }
  else if (scan)
  {
    usage(false);
    portScan();
  }
  else if (device.begin(portname))
  {
    static char path[PATH_MAX];
    snprintf(path, sizeof(path), "%s", argv[0]);
    char *p = strrchr(path, '.');
    if (!p)
      p = path + strlen(path);
    snprintf(p, sizeof(path) - strlen(path), "%s", ".chipid");
    device.setChipIDPath(path);

    Device::STATUS rv = device.program(speed);
    if (rv == Device::STATUS_OK)
    {
      usage(false);
      switch (speed)
      {
        case Device::SPEED_1M:   printf("Program Mode: 1M bps\n"); break;
        case Device::SPEED_500K: printf("Program Mode: 500K bps\n"); break;
        default:                 printf("Program Mode: %d bps\n", speed); break;
      }
      if (info)
        chipInfo();
      if (*savefile)
        device.readDevice(savefile, sector);
      if (erase)
        device.eraseDevice();
      if (*firmware)
        device.writeDevice(firmware, sector);
      device.restart();
    }
    else
      printf("Program Mode Failed = 0x%02X\n", rv);
    device.end();
  }
  else
  {
    usage(false);
    printf("Port Open Failed: \"%s\"\n", portname);
  }

  return 0;
}
